from setuptools import setup

setup(
    name="colorguard",
    version="0.2a2",
    packages=["colorguard"],
    url="http://github.com/reshanie/colorguard",
    license="MIT",
    author="Patrick Dill",
    author_email="jamespatrickdill@gmail.com",
    description="Better bit manipulation that allows indexing, and automatic bit flags.",
    requirements=[],
)
